#!/usr/bin/env python
# coding: utf-8
from jsoncomparekarate import compare

def long_line():
    print("-" * 120)


def run_tests():
    #keys must be str, int, float, bool or None, not bytes
    # a = {"姓名": "王大锤"}  # str and unicode (or bytes and str in python3) are compatible, useful in Chinese words...
    # b = {"姓名".encode("utf-8"): "王大锤".encode("utf-8")}
    # assert compare(a, b)
    # long_line()

    a = [[1, 2, 3], [4, 5, 6]]  # tuples and lists cannot be compared
    b = ([6, 5, 4], [3, 2, 1])
    assert not compare(a, b)
    long_line()

    a = [[1, 2, 3], [4, 5, 6]] # by default order is ignored, so these two are same
    b = [[3, 2, 1], [6, 5, 4]]
    assert compare(a, b)
    long_line()

    a = {"a": 1, "b": 3, "c": False, "d": "ok"}
    b = {"a": 1, "b": 2, "c": "False", "e": "ok"}  # False != "False"
    assert not compare(a, b)
    long_line()

    a = {"a": [1, {"k": ["ok"]}]}
    b = {"a": [1, {"k": ["error"]}]}  # ignoring list order, we aren't sure to pair {"k": ["ok"]} with {"k": ["error"]}
    assert not compare(a, b)
    long_line()

    a = {"a": [1, {"k": [0]}]}  # ignore path ["/a/1/k"] by specifying it as a list in the template
    b = {"a": [1, {"k": "#list"}]}
    assert compare(a, b)
    long_line()

    # a = [{"a": 1, "b": 2}, {"a": 5, "b": 4}]  # now we finally support regular expressions in ignore_path list
    # b = [{"a": 3, "b": 2}, {"a": 6, "b": 4}]  # in this case, only value of "b" concerned
    # check(a, b, ignore_list_seq=False, ignore_path=[r"^(/\d+/a)"])
    # long_line()

    # a = [{"a": 1, "b": 2}, {"a": 1, "b": 4}]  # also useful under list_seq ignored
    # b = [{"a": 2, "b": 4}, {"a": 2, "b": 2}]
    # check(a, b, ignore_path=[r"^(/\d+/a)"])
    # long_line()

    # a = [{"a": 1, "b": 3}, {"a": 1, "b": 4}]  # this time, 3 and 2 cannot match
    # b = [{"a": 2, "b": 4}, {"a": 2, "b": 2}]
    # assert not compare(a, b, ignore_path=[r"^(/\d+/a)"])
    # long_line()

    # a = [{"a": 1, "b": 2}, {"a": 3, "b": 4}, {"a": 5, "b": 4}]  # this time, only different frequency found
    # b = [{"a": 6, "b": 4}, {"a": 7, "b": 2}, {"a": 8, "b": 2}]  # but it will choose a random value of "a" to display
    # assert not compare(a, b, ignore_path=[r"^(/\d+/a)"]) # it's caused by logic restriction, don't get confused
    # long_line()

    a = {"a": [1, {"k": "#[_>0]", "l": "#bool"}, 3]}  # ignore two paths this time, "/a/1/k" and "/a/1/l"
    b = {"a": [1, {"k": [1], "l": False}, 3]}
    assert compare(a, b)
    long_line()

    a = '{"rtn": 0, "msg": "ok"}'  # can compare json string with python dict/list objects
    b = {"rtn": 1, "msg": "username not exist"}
    assert not compare(a, b)
    long_line()

    a = u'{"body":{"text":"你好"}}'  # both text and binary json strings are supported
    b = '{"body":{"text":"你好啊"}}'
    assert not compare(a, b)
    long_line()

    a = [1, 2, 2]  # even we ignore the order, the frequency of elements are concerned
    b = [1, 1, 2]
    assert not compare(a, b)
    long_line()

    a = [1, 2, 3]
    b = [1, 3, 4, 5]  # even if the length of lists are not equal, we can still know the difference
    assert not compare(a, b)
    long_line()

    a = [1, 2, 3]
    b = [1, 3, 4, 5]  # but we CANNOT keep the order of elements under different length even if ignore_list_seq is False
    assert not compare(a, b)
    long_line()

    a = [1.0]  # in face cp.compare(1, 1.0) is allowed, however non-standard jsons are not recommend
    b = [1]  # Integers and floats are compatible, including long of python 2
    assert compare(a, b)
    long_line()

    a = ["#regex ^(.*)$"]  # re-comparing enabled as default. Be careful bare r"^(.*)$" without list is considered as json-str
    b = ["anything"]  # use this to skip any unconcerned fields
    assert compare(a, b)
    long_line()

    a = ["#regex (.*)"]  # without ^-start or $-end, this won't be regarded as re-pattern
    b = ["anything"]
    assert compare(a, b)
    long_line()

    a = ["#regex ^(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2})$"]  # we can use re-comparing to confine formats but not values
    b = ["anything"]
    assert not compare(a, b)
    long_line()

    a = ["#regex ^(2019-07-01 \d{2}:\d{2}:\d{2})$"]  # e.g. this assertion will pass
    b = ["2019-07-01 12:13:14"]
    assert compare(a, b)
    long_line()

    a = ["#regex ^(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2})$", "#regex ^(.*)$"]
    b = ["anything", "otherthing"]  # when using re with order-ignored list, it will be crossing compare
    # be careful, potential chance of messy
    assert not compare(a, b)
    long_line()

    a = ["#regex ^(.*)$"]  # two re-pattern is not allowed
    b = ["#regex ^(.+)$"]
    try:
        compare(a, b)
    except Exception as e:
        print(e)
    else:
        raise AssertionError()
    long_line()

    a = ["#regex ^(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2})$", "otherthing"]
    b = ["anything", "#regex ^(.*)$"]  # this errors when comparing a[0] with b[1] due to the above rule
    try:
        compare(a, b)
    except Exception as e:
        print(e)
    else:
        raise AssertionError()
    long_line()

    a = '["^(2019-07-01 \\d{2}:\\d{2}:\\d{2})$"]'  # cannot compare a string and a list
    # or use '["^(2019-07-01 \\\\\d{2}:\\\\\d{2}:\\\\\d{2})$"]' will also work
    b = ["2019-07-01 12:13:14"]
    assert not compare(a, b)
    long_line()

    a = r'[r"^(2019-07-01 \d{2}:\d{2}:\d{2})$"]'  # cannot compare a string and a list
    b = ["2019-07-01 12:13:14"]
    assert not compare(a, b)
    long_line()

    a = ["#regex ^(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2})"]  # only fully match will pass re-comparing
    b = ["2019-07-01 12:13:14.567"]
    assert compare(a, b)
    long_line()

    a = ["#regex ^.*?(\d)-(\d)"]  # matching regex
    b = ["2019-07-01 12:13:14.567"]
    assert compare(a, b)
    long_line()

    a = [0.1 + 0.1 + 0.1]  # default we use accurate compare, since float compute causes accumulative errors
    b = [0.3]
    assert not compare(a, b)
    long_line()

    #check(a, b, ignore_list_seq=False, float_fuzzy_digits=6)  # so we can bear margin < 10e-6 now in float comparing
    # long_line()

    a = {"key": 1}
    b = {"key": 1.0}
    assert compare(a, b)
    long_line()

    #assert not compare(a, b)
    # long_line()

    a = {"a": {"b": 1, "c": 2, "d": 3}} # skip comparing "/a/b" and "/a/d"
    b = {"a": {"b": "##object", "c": 2, "d": "##object"}}
    assert compare(a, b)
    long_line()

    a = {"a": {"b": 1, "c": 2}}
    b = {"a": {"b": "##object", "c": 2, "d": 3}}
    # assert not compare(a, b, omit_path=["/a/b"])
    assert not compare(a, b)
    long_line()

    a = {"a": [{"b": 1, "c": 2}, {"d": 3, "c": 4}]} # omit paths "a/b" and "a/d"
    b = {"a": [{"b": "##object", "c": 2}, {"c": 4, "d": "##object"}]}
    assert compare(a, b)
    long_line()

    a = {"a": [{"b": 1, "c": 2}, {"c": 2}, {"d": 3, "c": 4}]}
    b = {"a": [{"b": "##object", "c": 2}, {"c": 4}, {"c": 4, "d": "##object"}]}
    assert not compare(a, b)
    long_line()

    a = { "data": "#regex unified-text:text-flow:text-max-chars.+" }
    b = { "data": "unified-text:text-flow:text-max-chars.plus-extra-chars" }
    assert compare(a, b)  # two strings with regular expression
    long_line()

    a = {"data": "#string"}
    b = {"data": "some-random-string"}
    assert compare(a, b)  # two strings with regular expression
    long_line()

    a = {"data": "#number"}
    b = {"data": 100}
    assert compare(a, b)  # two strings with regular expression
    long_line()

    a = {"data": "#object"}
    b = {"data": { "key1" : "value1", "key2" : "value2" }}
    assert compare(a, b)  # two strings with regular expression
    long_line()

    a = {"data": "#list"}
    b = {"data": ["item1", "item2"]}
    assert compare(a, b)  # two strings with regular expression
    long_line()

    a = {"data": "#[]"}
    b = {"data": ["item1", "item2"]}
    assert not compare(a, b)  # two strings with regular expression
    long_line()

    a = {"data": "#[]"}
    b = {"data": []}
    assert compare(a, b)  # two strings with regular expression
    long_line()

    a = {"data": "#[_ > 0]"}
    b = {"data": ["item1", "item2"]}
    assert compare(a, b)  # two strings with regular expression
    long_line()

    a = {"data": "#[_ > 0]"}
    b = {"data": []}
    assert not compare(a, b)  # two strings with regular expression
    long_line()

    a = {"data": "#[_ >= 0]"}
    b = {"data": []}
    assert compare(a, b)  # two strings with regular expression
    long_line()

    a = {"data": "#[]"}
    b = {"data": []}
    assert compare(a, b)  # two strings with regular expression
    long_line()

    a = {
       "id": "file",
       "value": "File",
       "popup": {
          "menuitem": [
             {"valueld1": "New", "onclickld1": "CreateNewDoc()"},
             {"valueld2": "Open"},
             {"valueld3": "Close", "onclickld3": "CloseDoc()"}
          ],
          "optionalItem": {}
       }
    }
    b = {
       "id": "#string",
       "value": "##string",
       "popup": {
          "menuitem": [
              {"valueld3": "Close", "onclickld3": "CloseDoc()"},
              {"valueld2": "Open", "onclickld2": "##string"},
              {"valueld1": "New", "onclickld1": "CreateNewDoc()"}
          ],
          "optionalItem": "#object",
          "notNeededItem": "##object"
       }
    }
    assert compare(a, b, strict_json_check = True)
    long_line()


if __name__ == "__main__":
    run_tests()



import json
import re

class _Compare(object):
    def __init__(self):
        self.is_in1_template = False
        self.is_in2_template = False
        self.supported_hash_types = ["#string", "#integer", "#float", "#boolean", "#bool", "#number", "#object", "#list",
                                     "#[", "#regex"]

    # Documentation
    # This function checks if the input is a valid JSON object. Returns True if the input is a valid JSON object, else
    # returns False.
    def is_valid_json_object(self, json_input):
        if isinstance(json_input, (dict, list)):
            try:
                # Try to serialize the object back to a JSON string
                json.dumps(json_input)
                return True  # If serialization is successful, it's a valid JSON object
            except (TypeError, ValueError):
                return False  # If serialization fails, it's not a valid JSON-like object
        return False  # If it's not a dictionary or list, it's not valid JSON-like object

    # Documentation
    # This function compares two JSONs. It takes two inputs, input_json1 and input_json2. It iterates over the keys of
    # input_json1 and compares the values of the keys in input_json1 with the corresponding keys in input_json2. If the
    # value of a key in input_json1 is a string and starts with "#", it is considered as a template value. The function
    # then compares the value of the key in input_json2 with the template value. If the value of a key in input_json1 is
    # a dictionary, the function recursively calls itself to compare the values of the keys in the dictionary. If the
    # value of a key in input_json1 is a list, the function compares the values of the elements in the list. If the
    # value of a key in input_json1 is a string, boolean, integer, or float, the function compares the value of the key
    # in input_json2 with the value of the key in input_json1. If the values are not equal, the function returns False.
    # If the values are equal, the function returns True.
    # Parameters
    # input_json1: dict (can be a template)
    # input_json2: dict
    def compare_jsons(self, input_json1, input_json2):
        for (k1, v1) in input_json1.items():
            #print(f"Starting comparison for key: {k1}, Expected Value: {v1}")

            # if the value is an optional value: specified by ##object, do nothing and continue
            if isinstance(v1, str) and v1.startswith("##"):
                continue

            # if the value is a string and starts with #, it is a template value so we need to compare it with the
            # corresponding value in input_json2
            if isinstance(v1, str) and v1.startswith("#"):

                #remove spaces and convert to lower case
                v1_no_space = v1.replace(" ", "").lower()

                if v1_no_space == "#string":
                    # if v1 is #string, in input_json2 ensure that v2 is a string; if not, return False, else continue
                    if not (k1 in input_json2 and isinstance(input_json2[k1], str)):
                        return False
                    else:
                        continue
                elif v1_no_space == "#integer":
                    # if v1 is #integer, in input_json2 ensure that v2 is a integer; if not, return False, else continue
                    if not (k1 in input_json2 and isinstance(input_json2[k1], int)):
                        return False
                    else:
                        continue
                elif v1_no_space == "#float":
                    # if v1 is #float, in input_json2 ensure that v2 is a float; if not, return False, else continue
                    if not (k1 in input_json2 and isinstance(input_json2[k1], float)):
                        return False
                    else:
                        continue
                elif v1_no_space == "#boolean" or v1 == "#bool":
                    # if v1 is #boolean, in input_json2 ensure that v2 is a boolean; if not, return False, else continue
                    if not (k1 in input_json2 and isinstance(input_json2[k1], bool)):
                        return False
                    else:
                        continue
                elif v1_no_space == "#number":
                    # if v1 is #number, in input_json2 ensure that v2 is a number; if not, return False, else continue
                    if not (k1 in input_json2 and (isinstance(input_json2[k1], int) or isinstance(input_json2[k1], float))):
                        return False
                    else:
                        continue
                elif v1_no_space == "#object":
                    # if v1 is #pbject, in input_json2 ensure that v2 is a dictionary; if not, return False, else continue
                    if not (k1 in input_json2 and isinstance(input_json2[k1], dict)):
                        return False
                    else:
                        continue
                elif v1_no_space == "#list":
                    # if v1 is #list, in input_json2 ensure that v2 is a list; if not, return False, else continue
                    if not (k1 in input_json2 and isinstance(input_json2[k1], list)):
                        return False
                    else:
                        continue
                elif v1_no_space.startswith("#["):
                    # if v1 starts with #[, in input_json2 ensure that v2 is a list of appropriate length; if not,
                    # return False, else continue
                    if v1_no_space == "#[]":
                        if not (k1 in input_json2 and isinstance(input_json2[k1], list) and len(input_json2[k1]) == 0):
                            return False
                        else:
                            continue
                    elif v1_no_space == "#[_>0]":
                        if not (k1 in input_json2 and isinstance(input_json2[k1], list) and len(input_json2[k1]) > 0):
                            return False
                        else:
                            continue
                    elif v1_no_space == "#[_>=0]":
                        if not (k1 in input_json2 and isinstance(input_json2[k1], list) and len(input_json2[k1]) >= 0):
                            return False
                        else:
                            continue
                    else:
                        print(f"Invalid list value found: {v1}")
                elif v1_no_space.startswith("#regex"):
                    # if v1 starts #regex, in input_json2 ensure that v2 matches the regex; if not, return False, else
                    # continue
                    #print(f"In the regex block: {v1}")
                    if v1_no_space.startswith("#regex"):
                        v1 = v1.replace("#regex ", "")
                    if not (k1 in input_json2):
                        return False
                    else:
                        v2 = input_json2[k1]
                        find = re.findall(v1, v2)
                        assert len(find) > 0, f"Cannot find the pattern in the string {v2}"
                        if not find:
                            print("re compare failed, empty match, see next line")
                            return False
                        if not find[0] == v2:
                            print(f"re compare failed, found {find[0]}, expect {v2}, see next line")
                            return False
                        continue
                else:
                    print(f"Invalid value found: {v1}")

            # if the value is string compare it with the corresponding value in input_json2 and continue if same, else
            # return False
            if isinstance(v1, str):
                if k1 in input_json2 and isinstance(input_json2[k1], str) and input_json1[k1] == input_json2[k1]:
                    continue
                else:
                    return False

            # if the value is boolean compare it with the corresponding value in input_json2 and continue if same, else
            # return False
            if isinstance(v1, bool):
                if k1 in input_json2 and isinstance(input_json2[k1], bool) and input_json1[k1] == input_json2[k1]:
                    continue
                else:
                    return False

            # if the value is int or float compare it with the corresponding value in input_json2 and continue if same,
            # else return False
            if isinstance(v1, int) or isinstance(v1, float):
                if k1 in input_json2 and (isinstance(input_json2[k1], int) or isinstance(input_json2[k1], float)) and input_json1[k1] == input_json2[k1]:
                    continue
                else:
                    return False

            # if the value is dictionary, iterate over all the key value pairs; return False if mismatch, continue if
            # all matches
            if isinstance(v1, dict):
                d1 = input_json1[k1]
                if k1 in input_json2 and isinstance(input_json2[k1], dict):
                    d2 = input_json2[k1]
                    return self.compare_jsons(d1, d2)

            # if the value is list, compare the elements of the list. If the lists are different, return False, else
            # continue
            if isinstance(v1, list):
                l1 = input_json1[k1]
                l2 = None
                if k1 in input_json2 and isinstance(input_json2[k1], list):
                    l2 = input_json2[k1]
                    if len(l1) != len(l2):
                        print(f"Lengths of lists for key {k1} are different: {l1}, {l2}")
                        return False
                return self.compare_lists(l1, l2)
        return True

    # Documentation
    # This function validates the input dictionaries for key value based comparison. It takes two inputs, input_dict1
    # and input_dict2. If any of the supported_hash_types are present in the input dictionaries, then they are marked as
    # template dictionary. The template dictionary has values indicating the type of value expected for a given key. The
    # values in the template dictionary can also indicate present or absence of a particular key. The function also
    # ensures that only one of the two input dictionaries is a template and throws an error if both inputs are
    # templates. Further it compares the two dictionaries and if the values are not equal, the function returns False.
    # If the values are equal, the function returns True.
    def compare_dicts(self, input_dict1, input_dict2):
        self.is_in1_template = False
        self.is_in2_template = False
        ijd1 = json.dumps(input_dict1)
        ijd2 = json.dumps(input_dict2)

        if any(x in ijd1 for x in self.supported_hash_types):
            self.is_in1_template = True
            #print("Hash found in input_json1")
        if any(x in ijd2 for x in self.supported_hash_types):
            self.is_in2_template = True
            #print("Hash found in input_json2")

        if self.is_in1_template and self.is_in2_template:
            print("Template is a JSON which has values referring to optional fields")
            raise Exception("Both inputs are templates. Only one input can be a template")

        if self.is_in2_template:
            #print("flipping the input dictionaries for internal comparison")
            return self.compare_jsons(input_dict2, input_dict1)
        else:
            return self.compare_jsons(input_dict1, input_dict2)

    # Documentation
    # This function compares the elements of two lists in order. If one of the elements in the list is a template or is
    # of the type list or dictionary, it will compare the elements of both lists accordingly.
    def compare_list_elements_in_order(self, input_list1, input_list2):
        for i, elem in enumerate(input_list1):
            if isinstance(elem, str) and elem.startswith("#regex"):
                a = elem.replace("#regex ", "")
                b = input_list2[i]
                find = re.findall(a, b)
                if not find:
                    return False
            elif isinstance(elem, dict):
                if not self.compare_jsons(elem, input_list2[i]):
                    return False
            elif isinstance(elem, list):
                if not self.compare_lists(elem, input_list2[i]):
                    return False
            else:
                if elem != input_list2[i]:
                    return False
        return True

    def compare_list_elements_out_of_order(self, input_list1, input_list2):
        if input_list1[0] and isinstance(input_list1[0], dict):
            ils1 = sorted(input_list1, key=lambda x: str(x))
            ils2 = sorted(input_list2, key=lambda x: str(x))
            return self.compare_list_elements_in_order(ils1, ils2)
        elif input_list1[0] and isinstance(input_list1[0], list):
            ils1 = sorted(input_list1)
            ils2 = sorted(input_list2)
            return self.compare_list_elements_in_order(ils1, ils2)
        else:
            ils1 = sorted(input_list1)
            ils2 = sorted(input_list2)
            return ils1 == ils2


    # Documentation
    # This function compares two lists, input_list1 and input_list2. It takes two inputs, input_list1 and input_list2.
    # Unless one of the lists is a template or in_order is set to True, the lists will be compared out of order. If all
    # the elements in the list are not of the same type, the function will do in order comparison.
    def compare_lists(self, input_list1, input_list2, in_order=False):
        #print(f"input_list1: {input_list1}, input_list2: {input_list2}")
        self.is_in1_template = False
        self.is_in2_template = False

        # check if all elements in the list are of the same type, if not then return False
        if len(input_list1) and not all(type(item) == type(input_list1[0]) for item in input_list1):
            in_order = True
        if len(input_list2) and not all(type(item) == type(input_list2[0]) for item in input_list2):
            in_order = True

        for elem in input_list1:
            if isinstance(elem, str) and any(x in elem for x in self.supported_hash_types):
                self.is_in1_template = True
                #print("Hash found in input_list1")
                break
        for elem in input_list2:
            if isinstance(elem, str) and any(x in elem for x in self.supported_hash_types):
                self.is_in2_template = True
                #print("Hash found in input_list2")
                break

        if self.is_in1_template and self.is_in2_template:
            print("Template is a List which has values referring to optional fields")
            raise Exception("Both inputs are templates. Only one input can be a template")

        if len(input_list1) != len(input_list2):
            print(f"Lengths of list1 ({len(input_list1)}) and list2 ({len(input_list2)}) are different")
            return False

        if self.is_in2_template:
            #print("flipping the input lists for internal comparison")
            input_list1, input_list2 = input_list2, input_list1

        if (self.is_in1_template or self.is_in2_template) or in_order:
            # one of the lists is a template and so we need to do in order comparison
            return self.compare_list_elements_in_order(input_list1, input_list2)
        else:
            # none of the lists are templates and so we can do out of order comparison
            return self.compare_list_elements_out_of_order(input_list1, input_list2)

    # Documentation
    # This function compares two inputs, in1 and in2. If the inputs are lists, the function calls the compare_lists
    # function. If the inputs are dictionaries, the function calls the compare_dicts function. If the inputs are not
    # lists or dictionaries, the function compares the inputs directly. If the inputs are equal, the function returns
    # True. If the inputs are not equal, the function returns False.
    def compare(self, in1, in2, strict_json_check):
        print(f"in1: {in1}, in1 type: {type(in1)}, in2: {in2}, int2 type: {type(in2)}")
        if strict_json_check:
            assert self.is_valid_json_object(in1)
            assert self.is_valid_json_object(in2)

        if isinstance(in1, list) and isinstance(in2, list):
            return self.compare_lists(in1, in2)
        elif isinstance(in1, dict) and isinstance(in2, dict):
            return self.compare_dicts(in1, in2)
        else:
            #print(f"{type(in1)} and {type(in2)} are not of type list or dict. Doing direct comparison..")
            return in1 == in2

def compare(a, b, strict_json_check=False):
    return _Compare().compare(a, b, strict_json_check)



#!/usr/bin/env python
# coding: utf-8
from distutils.core import setup
from os import path, chdir, system
from sys import argv

from setuptools import find_packages

if "upload" in argv:
    chdir("jsoncomparekarate")
    print("running test")
    assert system("python3 jsoncomparekaratetest.py") == 0
    chdir("..")

this_directory = path.abspath(path.dirname(__file__))

setup(
    name='json-compare-karate',
    version='1.0',
    packages=find_packages(),
    test_suite='jsoncomparekaratetest.py',
    author='Vin Pat',
    author_email='eatfrogfirst@outlook.com',
    description='A recursive json comparison library that handles karate style json comparison',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url='https://github.com/vinitatgithub/jsoncomparekarate',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    keywords='json comparison order karate optional fields',
    python_requires=">=3.5"
)
